﻿
; Levenshtein
; https://www.autoitscript.com/forum/topic/179886-comparing-strings/?tab=comments#comment-1291067


#include <File.au3>
#include <Math.au3>
#include <Array.au3>
#include <Timers.au3>



$corrcount = 0

$segfod = FileOpenDialog ("Select files containing segments", @ScriptDir, "All (*.*)")
$segfo = FileOpen ($segfod, 128)

$resultfsd = FileSaveDialog ("Select name of output file", @ScriptDir, "All (*.*)", 18, $segfod & "_output.txt")
$resultfo = FileOpen ($resultfsd, 129)

$similarsfsd = FileSaveDialog ("Select name of output file", @ScriptDir, "All (*.*)", 18, $resultfsd & "_similars.txt")
$similarsfo = FileOpen ($similarsfsd, 129)
 



$fuzzythreshold = InputBox ("", "Fuzzy threshold? 75 means '75% similarity or higher'. A lower number will result in more matches.", "75")

$lengththreshold = InputBox ("", "Length threshold? 1.0 means sentences must be identical; 0.75 means sentences outside of 75% of each other's length are not checked.", "0.75")
$spacelimit = InputBox ("", "Space limit? 5 means sentences of 4 or fewer spaces are not checked. Type 'N' to disable space limit.", "5")

$comparespacelimit = 0
$comparespacelimitreport = "No"
If $spacelimit <> "N" Then
$comparespacelimit = MsgBox (4, "Global space limit?", "Should we apply a space limit also to sentences that are being compared to the current sentence?", 0)
If $comparespacelimit = 6 Then ; yes
$comparespacelimitreport = "Yes"
Else
$comparespacelimitreport = "No"
EndIf
EndIf

$lengthlimit = InputBox ("", "Length limit? Sentences shorter than this are not checked. Type 'N' to disable length limit.", "10")
$upperlengthlimit = InputBox ("", "Upper length limit? Sentences LONGER than this are not checked. Type 'N' to disable length limit.", "N")
$methodused = "fuzzy trim, type 3 (levenshtein) v.3"

$segfr = FileRead ($segfo)

; The time stamps are for testing purposes only
$time1 = _Timer_Init ()
$time2 = "0"
$time3 = "0"

$listsplit = StringSplit ($segfr, @CRLF, 1)

$proccount = $listsplit[0]

For $outerloop = 1 to $listsplit[0] - 1

$firstline = $listsplit[$outerloop]

If $lengthlimit <> "N" Then
If StringLen ($firstline) < $lengthlimit Then
FileWrite ($resultfo, Round ((_Timer_Diff ($time1))/1000, 3) & @TAB & $firstline & @TAB & "(too short)" & @CRLF)
$proccount = $proccount - 1
ContinueLoop
EndIf
EndIf

If $upperlengthlimit <> "N" Then
If StringLen ($firstline) > $upperlengthlimit Then
FileWrite ($resultfo, Round ((_Timer_Diff ($time1))/1000, 3) & @TAB & $firstline & @TAB & "(too long)" & @CRLF)
$proccount = $proccount - 1
ContinueLoop
EndIf
EndIf

If $spacelimit <> "N" Then
$spacelimitcheck = StringReplace ($firstline, " ", " ")
If @extended < $spacelimit Then
FileWrite ($resultfo, Round ((_Timer_Diff ($time1))/1000, 3) & @TAB & $firstline & @TAB & "(too few spaces)" & @CRLF)
$proccount = $proccount - 1
ContinueLoop
EndIf
EndIf




; If $outerloop = 1 Then
; ToolTip ("Now processing line " & $outerloop & " of approx " & $proccount & " remaining" & @CRLF & $firstline, 200, 200)
; Else
ToolTip ("Now processing line " & $outerloop & "; Approx " & $proccount & " lines remaining; " & $corrcount & " matches found so far." & @CRLF & @CRLF & Round ((_Timer_Diff ($time1))/1000, 3) & " seconds since start" & @CRLF & Round ((_Timer_Diff ($time2))/1000, 3) & " seconds since previous segment" & @CRLF & Round (((_Timer_Diff ($time1))/1000) / $outerloop, 3) & " average seconds per segment so far", 200, 200)
; EndIf



If StringInStr ($listsplit[$outerloop], "::!!!::") Then
$proccount = $proccount - 1
ContinueLoop
Else
FileWrite ($resultfo, Round ((_Timer_Diff ($time1))/1000, 3) & @TAB & $firstline & @TAB & "(no match)" & @CRLF)
$proccount = $proccount - 1
EndIf

$time2 = _Timer_Init ()






For $innerloop = $outerloop + 1 to $listsplit[0]

$otherline = $listsplit[$innerloop]

If StringInStr ($otherline, "::!!!::") Then
ContinueLoop
EndIf

If $comparespacelimit = 6 Then
$spacelimitcheck = StringReplace ($otherline, " ", " ")
If @extended > ($spacelimit / $lengththreshold) Then
ContinueLoop
EndIf
EndIf

If StringLen ($otherline) < ($lengthlimit * $lengththreshold) Then
ContinueLoop
EndIf

If StringLen ($firstline) / StringLen ($otherline) < $lengththreshold Then
ContinueLoop
EndIf
If StringLen ($firstline) / StringLen ($otherline) > (1 / $lengththreshold) Then
ContinueLoop
EndIf







$simsim = Int (altLevenshtein($firstline, $otherline))

If $simsim > $fuzzythreshold Then
FileWrite ($resultfo, Int ((_Timer_Diff ($time1))/1000) & @TAB & "{{" & $otherline & "}}" & @TAB & $simsim & @CRLF)
FileWrite ($similarsfo, $firstline & @TAB & $simsim & @TAB & $otherline & @CRLF)
$listsplit[$innerloop] = $listsplit[$innerloop] & "::!!!::"
$corrcount = $corrcount + 1
$proccount = $proccount - 1
EndIf

Next

; $time3 = _Timer_Init ()

Next

ToolTip ("")

$finalreport = "" 
$finalreport = $finalreport & "=========|=========|=========|=========|=========" & @CRLF
$finalreport = $finalreport & @CRLF 
$finalreport = $finalreport & "Input file: " & StringReplace ($segfod, @ScriptDir & "\", "") & @CRLF
$finalreport = $finalreport & "Output file: " & StringReplace ($resultfsd, @ScriptDir & "\", "") & @CRLF
$finalreport = $finalreport & "Fuzzy threshold: " & $fuzzythreshold & "%" & @CRLF
$finalreport = $finalreport & "Length threshold: " & $lengththreshold & @CRLF
$finalreport = $finalreport & "Space limit: " & $spacelimit & @CRLF
$finalreport = $finalreport & "Global space limit: " & $comparespacelimitreport & @CRLF
$finalreport = $finalreport & "Length limit: " & $lengthlimit & @CRLF
$finalreport = $finalreport & "Upper length limit: " & $upperlengthlimit & @CRLF
$finalreport = $finalreport & "Method used: " & $methodused & @CRLF
$finalreport = $finalreport & @CRLF 
$finalreport = $finalreport & "Total time: " & Round ((_Timer_Diff ($time1))/1000, 3) & " seconds" & @CRLF
$finalreport = $finalreport & "# of segments: " & $listsplit[0] & @CRLF
$finalreport = $finalreport & "# of matches: " & $corrcount & @CRLF
$finalreport = $finalreport & @CRLF 
$finalreport = $finalreport & "=========|=========|=========|=========|=========" & @CRLF

FileWrite ($resultfo, @CRLF & @CRLF & @CRLF & $finalreport & @CRLF & @CRLF & @CRLF)

MsgBox (0, "Done!", $finalreport, 0)

Exit


; ========================================================================================================


; ========================================================================================================

; The third-party function also has the ability to convert to percentages, but
; we don't do that, for being able to compare it with other "Levenshtein"s.

; Local $altLevenshtein = altLevenshtein($s1, $s2)
; Local $iMaxLen = StringLen (  StringLen($s1) > StringLen($s2) ? $s1 : $s2 )
; Local $percent = Round(($iMaxLen - $Levenshtein) * 100 / $iMaxLen, 2)
; ConsoleWrite($percent & "%")

Func altLevenshtein( $a, $b )
    Local $i, $j, $cost, $d[1], $min1, $min2, $min3

    If StringLen( $a ) = 0 Then Return StringLen( $b )
    If StringLen( $b ) = 0 Then Return StringLen( $a )

    ReDim $d[ StringLen( $a ) + 1][ StringLen( $b ) + 1]

    For $i = 0 To StringLen( $a )
        $d[$i][0] = $i
    Next

    For $j = 0 To StringLen( $b )
        $d[ 0][$j] = $j
    Next

    For $i = 1 To StringLen( $a )
        For $j = 1 To StringLen( $b )
            $cost =  ( StringMid($a, $i, 1) = StringMid($b, $j, 1) ? 0 : 1)

            $min1 = $d[$i - 1][$j] + 1
            $min2 = $d[$i][$j - 1] + 1
            $min3 = $d[$i - 1][$j - 1] + $cost

            If $min1 <= $min2 And $min1 <= $min3 Then
                $d[$i][$j] = $min1
            ElseIf $min2 <= $min1 And $min2 <= $min3 Then
                $d[$i][$j] = $min2
            Else
                $d[$i][$j] = $min3
            EndIf
        Next
    Next

    Return $d[StringLen( $a )][StringLen( $b )]
EndFunc