﻿
; Levenshtein
; https://www.autoitscript.com/forum/topic/144812-levenshtein-distance-function-problem/?tab=comments#comment-1022120

#include <File.au3>
#include <Math.au3>
#include <Array.au3>
#include <Timers.au3>

$corrcount = 0

$segfod = FileOpenDialog ("Select files containing segments", @ScriptDir, "All (*.*)")
$segfo = FileOpen ($segfod, 128)

$resultfsd = FileSaveDialog ("Select name of output file", @ScriptDir, "All (*.*)", 18, $segfod & "_output.txt")
$resultfo = FileOpen ($resultfsd, 129)

$similarsfsd = FileSaveDialog ("Select name of output file", @ScriptDir, "All (*.*)", 18, $resultfsd & "_similars.txt")
$similarsfo = FileOpen ($similarsfsd, 129)
 




$fuzzythreshold = InputBox ("", "Fuzzy threshold? 75 means '75% similarity or higher'. A lower number will result in more matches.", "75")

$lengththreshold = InputBox ("", "Length threshold? 1.0 means sentences must be identical; 0.75 means sentences outside of 75% of each other's length are not checked.", "0.75")
$spacelimit = InputBox ("", "Space limit? 5 means sentences of 4 or fewer spaces are not checked. Type 'N' to disable space limit.", "5")

$comparespacelimit = 0
$comparespacelimitreport = "No"
If $spacelimit <> "N" Then
$comparespacelimit = MsgBox (4, "Global space limit?", "Should we apply a space limit also to sentences that are being compared to the current sentence?", 0)
If $comparespacelimit = 6 Then ; yes
$comparespacelimitreport = "Yes"
Else
$comparespacelimitreport = "No"
EndIf
EndIf

$lengthlimit = InputBox ("", "Length limit? Sentences shorter than this are not checked. Type 'N' to disable length limit.", "10")
$upperlengthlimit = InputBox ("", "Upper length limit? Sentences LONGER than this are not checked. Type 'N' to disable length limit.", "N")
$methodused = "fuzzy trim, type 4 (levenshtein) v.3"

$segfr = FileRead ($segfo)

; The time stamps are for testing purposes only
$time1 = _Timer_Init ()
$time2 = "0"
$time3 = "0"

$listsplit = StringSplit ($segfr, @CRLF, 1)

$proccount = $listsplit[0]

For $outerloop = 1 to $listsplit[0] - 1

$firstline = $listsplit[$outerloop]

If $lengthlimit <> "N" Then
If StringLen ($firstline) < $lengthlimit Then
FileWrite ($resultfo, Round ((_Timer_Diff ($time1))/1000, 3) & @TAB & $firstline & @TAB & "(too short)" & @CRLF)
$proccount = $proccount - 1
ContinueLoop
EndIf
EndIf

If $upperlengthlimit <> "N" Then
If StringLen ($firstline) > $upperlengthlimit Then
FileWrite ($resultfo, Round ((_Timer_Diff ($time1))/1000, 3) & @TAB & $firstline & @TAB & "(too long)" & @CRLF)
$proccount = $proccount - 1
ContinueLoop
EndIf
EndIf

If $spacelimit <> "N" Then
$spacelimitcheck = StringReplace ($firstline, " ", " ")
If @extended < $spacelimit Then
FileWrite ($resultfo, Round ((_Timer_Diff ($time1))/1000, 3) & @TAB & $firstline & @TAB & "(too few spaces)" & @CRLF)
$proccount = $proccount - 1
ContinueLoop
EndIf
EndIf




; If $outerloop = 1 Then
; ToolTip ("Now processing line " & $outerloop & " of approx " & $proccount & " remaining" & @CRLF & $firstline, 200, 200)
; Else
ToolTip ("Now processing line " & $outerloop & "; Approx " & $proccount & " lines remaining; " & $corrcount & " matches found so far." & @CRLF & @CRLF & Round ((_Timer_Diff ($time1))/1000, 3) & " seconds since start" & @CRLF & Round ((_Timer_Diff ($time2))/1000, 3) & " seconds since previous segment" & @CRLF & Round (((_Timer_Diff ($time1))/1000) / $outerloop, 3) & " average seconds per segment so far", 200, 200)
; EndIf


If StringInStr ($listsplit[$outerloop], "::!!!::") Then
$proccount = $proccount - 1
ContinueLoop
Else
FileWrite ($resultfo, Round ((_Timer_Diff ($time1))/1000, 3) & @TAB & $firstline & @TAB & "(no match)" & @CRLF)
$proccount = $proccount - 1
EndIf

$time2 = _Timer_Init ()






For $innerloop = $outerloop + 1 to $listsplit[0]

$otherline = $listsplit[$innerloop]

If StringInStr ($otherline, "::!!!::") Then
ContinueLoop
EndIf

If $comparespacelimit = 6 Then
$spacelimitcheck = StringReplace ($otherline, " ", " ")
If @extended > ($spacelimit / $lengththreshold) Then
ContinueLoop
EndIf
EndIf

If StringLen ($otherline) < ($lengthlimit * $lengththreshold) Then
ContinueLoop
EndIf

If StringLen ($firstline) / StringLen ($otherline) < $lengththreshold Then
ContinueLoop
EndIf
If StringLen ($firstline) / StringLen ($otherline) > (1 / $lengththreshold) Then
ContinueLoop
EndIf







$simsim = Int (_LevDis($firstline, $otherline))

If $simsim > $fuzzythreshold Then
FileWrite ($resultfo, Int ((_Timer_Diff ($time1))/1000) & @TAB & "{{" & $otherline & "}}" & @TAB & $simsim & @CRLF)
FileWrite ($similarsfo, $firstline & @TAB & $simsim & @TAB & $otherline & @CRLF)
$listsplit[$innerloop] = $listsplit[$innerloop] & "::!!!::"
$corrcount = $corrcount + 1
$proccount = $proccount - 1
EndIf

Next

; $time3 = _Timer_Init ()

Next

ToolTip ("")

$finalreport = "" 
$finalreport = $finalreport & "=========|=========|=========|=========|=========" & @CRLF
$finalreport = $finalreport & @CRLF 
$finalreport = $finalreport & "Input file: " & StringReplace ($segfod, @ScriptDir & "\", "") & @CRLF
$finalreport = $finalreport & "Output file: " & StringReplace ($resultfsd, @ScriptDir & "\", "") & @CRLF
$finalreport = $finalreport & "Fuzzy threshold: " & $fuzzythreshold & "%" & @CRLF
$finalreport = $finalreport & "Length threshold: " & $lengththreshold & @CRLF
$finalreport = $finalreport & "Space limit: " & $spacelimit & @CRLF
$finalreport = $finalreport & "Global space limit: " & $comparespacelimitreport & @CRLF
$finalreport = $finalreport & "Length limit: " & $lengthlimit & @CRLF
$finalreport = $finalreport & "Upper length limit: " & $upperlengthlimit & @CRLF
$finalreport = $finalreport & "Method used: " & $methodused & @CRLF
$finalreport = $finalreport & @CRLF 
$finalreport = $finalreport & "Total time: " & Round ((_Timer_Diff ($time1))/1000, 3) & " seconds" & @CRLF
$finalreport = $finalreport & "# of segments: " & $listsplit[0] & @CRLF
$finalreport = $finalreport & "# of matches: " & $corrcount & @CRLF
$finalreport = $finalreport & @CRLF 
$finalreport = $finalreport & "=========|=========|=========|=========|=========" & @CRLF

FileWrite ($resultfo, @CRLF & @CRLF & @CRLF & $finalreport & @CRLF & @CRLF & @CRLF)

MsgBox (0, "Done!", $finalreport, 0)

Exit


; ========================================================================================================


Func _LevDis($s, $t)
Local $m, $n, $iMaxM, $iMaxN

$n = StringLen($s)
$m = StringLen($t)
$ss = StringSplit($s, "")
$tt = StringSplit($t, "")
$iMaxN = $n + 1
$iMaxM = $m + 1
Dim $d[$iMaxN + 1][$iMaxM + 1]
$d[0][0] = 0

If $n = 0 Then
Return $m
ElseIf $m = 0 Then
Return $n
EndIf

For $i = 1 To $n
$d[$i][0] = $d[$i - 1][0] + 1
Next
For $j = 1 To $m
$d[0][$j] = $d[0][$j - 1] + 1
Next

For $i = 1 To $n
For $j = 1 To $m
$jj = $j - 1
$ii = $i - 1
Local $cost
If (StringMid($s, $i, 1) = StringMid($t, $j, 1)) Then
$cost = 0
Else
$cost = 1
EndIf
$d[$i][$j] = _Min(_Min($d[$ii][$j] + 1, _
$d[$i][$jj] + 1), $d[$ii][$jj] + $cost)
Next
Next
Return $d[$n][$m]
EndFunc
